/*
 * httptime.cpp
 * blackcore networking
 *
 * source@source.blackcore.net
 *
 * Application measures time between an HTTP request and the server's
 * response to determine how lagged the remote server is lagged.
 *
 *
 * connect_time: time it took for connect() to open the socket
 * server_response: time from send() and first recv()
 * document_time: time from send() and first recv()
 * document_size: total size of recv()
 * document_rate: estimate of how fast document transfered
 *
 * linux: c++ httptime.cpp -o httptime
 * Windows:
 *  Visual Studio (run vsvars32.bat): cl httptime.cpp ws2_32.lib -DWIN32
 *  Dev-C++: c++ httptime.cpp -o httptime.exe -lws32_32
*/
#include <stdio.h>
#include <string.h>
#ifdef WIN32
 #include <winsock2.h>
 #include <windows.h>
 #define snprintf _snprintf
 #define vsnprintf _vsnprintf
 typedef LONGLONG longlong_t;
 #pragma comment(lib, "ws2_32.lib")
#else
 #include <errno.h>
 #include <stdarg.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <sys/socket.h>
 #include <netdb.h>
 #include <stdlib.h>
 #include <sys/time.h>
 extern int h_errno;
 extern int errno;
 typedef int SOCKET;
 typedef long long longlong_t;
#endif /* WIN32 */
typedef int ssize_t;
#define FLAGS_HTTP_SHOW_STATUS 0x1
int debug_level = 0;
int flags = 0;
char *convertbw(unsigned long);
char *translate_http_code(char *);
void sockerr(char *);
long getaddress(char *);
void cleansocket(SOCKET);
longlong_t getnow(void);
int debug(char *, ...);
int geturlinfo(char *,char *,int,char *,int);
void usage(char *);
int main(int argc, char *argv[]) {
	SOCKET s;
	int n,  total, i, opt, opti;
	longlong_t begin, end;
	char data[256], uri[128], host[64], path[128], port[8];
	struct sockaddr_in peer;
	if (argc  < 2) {
		usage(argv[0]);
		return 1;
	}

	memset(data, 0, sizeof(data));

	for(i=1;i<argc;i++) {
		strncat(data, argv[i], sizeof(data));
		if (i+1 < argc) strncat(data, " ", sizeof(data));
	}
	n = strlen(data);
	memset(uri, '\0', sizeof(uri));
	memset(host, '\0', sizeof(host));
	memset(path, '\0', sizeof(path));
	memset(port, '\0', sizeof(port));
/*
 * I don't know how to use getopt! Does Windows even have a getopt()
 * call?
*/
	for(i=0,opt=0;i < n; i++) {
		if (opt && data[i] == ' ') opt=0;
		else if (opt) {
			switch(opt) {
				case 'p':
					port[opti++] = data[i];
				break;
				case 'w':
					if (data[i] == '/') continue;
					path[opti++] = data[i];
				break;
				case 'h':
				case 'H':
					host[opti++] = data[i];
				break;
				default:
				break;
			}
		} else if (data[i] == '-' && i+1 < n) {
			opt = data[i+1];
			switch(opt) {
				case 's':
					flags |= FLAGS_HTTP_SHOW_STATUS;
				break;
				case '?':
					usage(argv[0]);
					return 1;
				break;
				case 'd':
					debug_level++;
				break;
				default:
				break;
			}
			debug("opt: %c", opt);
			opti=0;
			i+=2;
		}
		else if (data[i] != ' ') {
			char byte[2];
			byte[0] = data[i];
			byte[1]=0;
			strncat(uri, byte, sizeof(uri));
		}
	}
	if (uri[0])
		debug("uri_info: %s", uri);
	if (host[0])
		debug("set_host: %s", host);
	if (port[0])
		debug("set_port: %s", port);
	debug("debug_level: %d", debug_level);
	if (port[0] == 0)
		peer.sin_port = htons(80);
	else {
		sscanf(port, "%d", &peer.sin_port);
		peer.sin_port = htons(peer.sin_port);
	}
	if (host[0] == '\0') {
		if (geturlinfo(uri, host, sizeof(host), path, sizeof(path)) == -1) {
			fprintf(stderr,"No hostname specified!\n");
			return 1;
		}
	}
#ifdef WIN32
	WSADATA Data;
	WSAStartup(MAKEWORD(2,0), &Data);
#endif

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		sockerr("socket");
		return 1;
	}

	peer.sin_family = AF_INET;
	peer.sin_addr.s_addr = getaddress(host);

	if (peer.sin_addr.s_addr == 1) {
		printf("Could not resolve hostname! (%s)\n", host);
		cleansocket(s);
		return 1;
	}
	begin = getnow();
	debug("connect: %s:%d", inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));
	if (connect(s, (struct sockaddr *)&peer, sizeof(peer)) < 0) {
		sockerr("connect");
		return 1;
	}
	begin = getnow() - begin;
	printf("connect_time: %dms\n", begin);
	n = snprintf(data, sizeof(data), "GET /%s HTTP/1.0\r\nHost: %s\r\nUser-Agent: httptime client\r\nConnection: Close\r\n\r\n", path, host);

	debug("send: host=%s path=/%s", host, path);
	begin = getnow();
	if (send(s, data, n, 0) < 0) {
		sockerr("send");
		cleansocket(s);
		return 1;
	}
	total = 0;
	debug("recv: waiting on data");
	while ((n = recv(s, data, sizeof(data), 0))) {
		if (!total) {
			if (flags & FLAGS_HTTP_SHOW_STATUS) {
				char *a;
				a = strtok(data, " ");
				if (a)
					a = strtok(NULL, " ");
				if (a)
					printf("http_code: %s %s\n", a, translate_http_code(a));
				
			}
			printf("server_response: %dms\n", (getnow() - begin));
		}
		total+=n;
	}
	end = (getnow() - begin);
	printf("document_time: %dms\n", end);
	printf("document_size: %s\n", convertbw(total));
	printf("document_rate: %s/s\n", convertbw((unsigned long)((end / 1000)?total/(end / 1000):total)));
	cleansocket(s);
	return 0;
}
void cleansocket(SOCKET s) {
#ifdef WIN32
	closesocket(s);
#else
	shutdown(s, SHUT_RDWR);
#endif /* WIN32 */
}
void sockerr(char *func) {
#ifdef WIN32
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError(); 

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

	printf("%s: %s", func, lpMsgBuf); 

    LocalFree(lpMsgBuf);
    ExitProcess(dw); 
#else
	fprintf(stderr, "%s: %s\n", func, strerror(errno));
#endif /* WIN32 */
	return;
}
long getaddress(char *input) {
	struct hostent *he;
	struct in_addr ret;

	debug("getaddress: %s", input);
	if (inet_addr(input) == -1) {
		he = gethostbyname(input);
#ifdef WIN32
		if (WSAGetLastError() != 0)
			return 1;
#else
		if (!(he = gethostbyname(input)))
			return 1;
#endif /* WIN32 */
		memcpy(&ret, he->h_addr_list[0], sizeof(ret));
		return ret.s_addr;
	}
	return inet_addr(input);
}
longlong_t getnow() {
#ifdef WIN32
	longlong_t ticks = (longlong_t)GetTickCount();
#else
	struct timeval tv;
	long long ticks;
	gettimeofday(&tv, NULL);
	ticks = (long long) (tv.tv_sec * 1000) + (tv.tv_usec / 1000);
#endif
	return ticks;
}
char *convertbw(unsigned long input) {
	char a[4], string[16];
	float length;
	memset(a, '\0', sizeof(a));
	if (input > 1073741824) {
		length = (float)input / 1073741824;
		strcpy(a, "GB");
	} else if (input > 1048576) {
		length = (float)input / 1048576;
		strcpy(a, "MB");
	} else if (input > 1024) {
		length = (float)input / 1024;
		strcpy(a, "KB");
	} else
		length = (float)input;
	snprintf(string, sizeof(string), "%.2f%s", length, a[0]?a:" bytes");
	return strdup(string);
}
int debug(char *format, ...) {
	char buf[512];
	int n;
	va_list vargs;
	va_start(vargs, format);
	n = vsnprintf(buf, 512, format, vargs);
	if (debug_level)
		fprintf(stderr, "(debug) %s\n", buf);
	va_end(vargs);
	return n;
}
int geturlinfo(char *uri, char *host, ssize_t host_size, char *path, ssize_t path_size) {
	int n = 0;
	int i, t, a, b;
	int uri_length = strlen(uri);
	memset(host, '\0', host_size);
	memset(path, '\0', path_size);
	if (uri_length > 6)
		if (uri[4] == ':') n = 7;
	for(i=n, t=0, a=0, b=0;i<uri_length;i++) {
		if (uri[i] == '/') t++;
		else if (t) path[a++]=uri[i];
		else
			host[b++]=uri[i];
	}

	if (host[0] == '\0') return -1;
	return 0;
}
void usage(char *input) {
	fprintf(stderr,"usage: %s <options> [host]\n\n -h <hostname>\n -p port\n -w <path>\n -s show http response code\n\n", input);
}
char *translate_http_code(char *input) {
	if (!strcmp(input, "200")) return strdup("OK");
	if (!strcmp(input, "404")) return strdup("Not found");
	if (!strcmp(input, "500")) return strdup("Internal server error");
	return strdup("Undefined");
}
